@@ -8,8 +8,8 @@ from django.conf import settings
 from django.contrib.auth import get_user_model
 from django.core.cache import cache
 from django.core.exceptions import ValidationError
-from django.core.validators import MaxLengthValidator, MaxValueValidator, \
-    MinLengthValidator, MinValueValidator, URLValidator
+from django.core.validators import MaxLengthValidator, MaxValueValidator, MinLengthValidator, \
+    MinValueValidator, URLValidator
 from django.db import models
 from django.db.models import OuterRef, Subquery
 from django.utils.timezone import now
