@@ -8,60 +8,113 @@ from decimal import Decimal
 
 from django.conf import settings
 from django.core.cache import cache
+from django.db import transaction
 from django.db.models import Count, Max, OuterRef, Q, Subquery
 from django.db.models.functions import Substr
-from django.utils.timezone import make_aware
+from django.utils.timezone import make_aware, now
 from influxdb import InfluxDBClient
 from satnogsdecoders import __version__ as decoders_version
 from satnogsdecoders import decoder
 
-from db.base.models import DemodData, LatestTleSet, Mode, Satellite, \
-    Telemetry, Tle, Transmitter
+from db.base.models import DemodData, LatestTleSet, Mode, Satellite, Tle, Transmitter
 
 LOGGER = logging.getLogger('db')
 
 
-def remove_latest_tle_set(satellite_pk):
-    """Remove LatestTleSet for specific Satellite"""
-    LatestTleSet.objects.filter(satellite__pk=satellite_pk).delete()
+def remove_latest_tle_set(satellite_id):
+    """Remove LatestTleSet entry for specific Satellite"""
+    LatestTleSet.objects.filter(satellite__pk=satellite_id).delete()
 
 
-def update_latest_tle_sets(satellites=None):
-    """Update LatestTleSet for specific or all Satellites"""
-    if not satellites:
+def update_latest_tle_sets(satellite_ids=None):
+    """Update LatestTleSet entries for all or specific Satellites"""
+    # Select satellite models
+    if satellite_ids:
+        satellites = Satellite.objects.filter(pk__in=satellite_ids)
+    else:
         satellites = Satellite.objects.exclude(status__in=['re-entered', 'future'])
 
-    sub_subquery = Tle.objects.filter(satellite__pk__in=satellites).filter(
-        satellite=OuterRef('satellite'), tle_source=OuterRef('tle_source')
-    ).order_by('-updated')
-    subquery = Tle.objects.filter(pk=Subquery(sub_subquery.values('pk')[:1])
-                                  ).filter(satellite=OuterRef('satellite')
-                                           ).annotate(epoch=Max(Substr('tle1', 19, 14))
-                                                      ).order_by('-epoch')
-    tle_sets = Tle.objects.filter(pk=Subquery(subquery.values('pk')[:1]))
+    # Create dictionary with Satellite ids as keys and Tle ids as values
+    latest_triplets = LatestTleSet.objects.filter(
+        satellite__in=satellites
+    ).values_list('satellite', 'latest', 'latest_distributable')
+    latest_dictionary = {
+        latest_triplet[0]: (latest_triplet[1], latest_triplet[2])
+        for latest_triplet in latest_triplets
+    }
 
-    sub_subquery = Tle.objects.filter(satellite__pk__in=satellites).filter(
-        tle_source__in=settings.TLE_SOURCES_REDISTRIBUTABLE
-    ).filter(
-        satellite=OuterRef('satellite'), tle_source=OuterRef('tle_source')
-    ).order_by('-updated')
-    subquery = Tle.objects.filter(pk=Subquery(sub_subquery.values('pk')[:1])
-                                  ).filter(satellite=OuterRef('satellite')
-                                           ).annotate(epoch=Max(Substr('tle1', 19, 14))
-                                                      ).order_by('-epoch')
-    distributable_tle_sets = Tle.objects.filter(pk=Subquery(subquery.values('pk')[:1]))
+    # For each satellite update LatestTleSet
+    for satellite in satellites:
 
-    for tle_set in tle_sets:
-        latest_tle_set, _ = LatestTleSet.objects.get_or_create(satellite=tle_set.satellite)
-        latest_tle_set.latest = tle_set
-        latest_tle_set.save(update_fields=['latest'])
-    for distributable_tle_set in distributable_tle_sets:
-        latest_tle_set, _ = LatestTleSet.objects.get_or_create(
-            satellite=distributable_tle_set.satellite
-        )
-        latest_tle_set.latest_distributable = distributable_tle_set
-        latest_tle_set.save(update_fields=['latest_distributable'])
+        # Keep the Tle ids of the LatestTleSet to check Tle entries inserted after them
+        # If there isn't one (new satellite or remove Tle entry) then check all of them
+        tle_id, tle_id_dist = (0, 0)
+        if satellite.id in latest_dictionary:
+            tle_ids = latest_dictionary[satellite.id]
+            tle_id = tle_ids[0] if tle_ids[0] else 0
+            tle_id_dist = tle_ids[1] if tle_ids[1] else 0
 
+        # Query for the latest Tle set for this satellite
+        sub_subquery = Tle.objects.filter(
+            pk__gte=tle_id, satellite=satellite
+        ).exclude(tle_source__in=settings.TLE_SOURCES_IGNORE_FROM_LATEST).filter(
+            satellite=OuterRef('satellite'), tle_source=OuterRef('tle_source')
+        ).order_by('-updated')
+        subquery = Tle.objects.filter(
+            pk__gte=tle_id, satellite=satellite
+        ).exclude(tle_source__in=settings.TLE_SOURCES_IGNORE_FROM_LATEST
+                  ).filter(pk=Subquery(sub_subquery.values('pk')[:1])
+                           ).filter(satellite=OuterRef('satellite')
+                                    ).annotate(epoch=Max(Substr('tle1', 19, 14))
+                                               ).order_by('-epoch')
+        new_latest = Tle.objects.filter(
+            pk__gte=tle_id, satellite=satellite
+        ).exclude(tle_source__in=settings.TLE_SOURCES_IGNORE_FROM_LATEST
+                  ).filter(pk=Subquery(subquery.values('pk')[:1]))
+
+        # Query for the latest Tle set that is distributable for this satellite
+        sub_subquery = Tle.objects.filter(
+            pk__gte=tle_id_dist,
+            satellite=satellite,
+            tle_source__in=settings.TLE_SOURCES_REDISTRIBUTABLE
+        ).exclude(tle_source__in=settings.TLE_SOURCES_IGNORE_FROM_LATEST).filter(
+            satellite=OuterRef('satellite'), tle_source=OuterRef('tle_source')
+        ).order_by('-updated')
+        subquery = Tle.objects.filter(
+            pk__gte=tle_id_dist,
+            satellite=satellite,
+            tle_source__in=settings.TLE_SOURCES_REDISTRIBUTABLE
+        ).exclude(tle_source__in=settings.TLE_SOURCES_IGNORE_FROM_LATEST
+                  ).filter(pk=Subquery(sub_subquery.values('pk')[:1])
+                           ).filter(satellite=OuterRef('satellite')
+                                    ).annotate(epoch=Max(Substr('tle1', 19, 14))
+                                               ).order_by('-epoch')
+        new_latest_dist = Tle.objects.filter(
+            pk__gte=tle_id_dist,
+            satellite=satellite,
+            tle_source__in=settings.TLE_SOURCES_REDISTRIBUTABLE
+        ).exclude(tle_source__in=settings.TLE_SOURCES_IGNORE_FROM_LATEST
+                  ).filter(pk=Subquery(subquery.values('pk')[:1]))
+
+        # Add the latest Tle set if there is a LatestTleSet entry, if not create one
+        if new_latest:
+            LatestTleSet.objects.update_or_create(
+                satellite=satellite, defaults={
+                    'latest': new_latest[0],
+                    'last_modified': now()
+                }
+            )
+        # Add the latest distributable Tle set if there is a LatestTleSet entry, if not create one
+        if new_latest_dist:
+            LatestTleSet.objects.update_or_create(
+                satellite=satellite,
+                defaults={
+                    'latest_distributable': new_latest_dist[0],
+                    'last_modified': now()
+                }
+            )
+
+    # Remove any LatestTleSet that hasn't any Tle entry (satellite without Tle entries)
     LatestTleSet.objects.filter(latest__isnull=True, latest_distributable__isnull=True).delete()
 
 
@@ -286,80 +339,72 @@ def write_influx(json_obj):
     client.write_points(json_obj)
 
 
-def decode_data(norad, period=None, handle=None):  # pylint: disable=R0912,R0914
+def decode_demoddata(demoddata, satellite, tlmdecoder):
+    """Decode a DemodData object and push it either in InfluxDB instance or in local DB"""
+    try:
+        decoder_class = getattr(decoder, tlmdecoder.decoder.capitalize())
+    except AttributeError:
+        return
+    try:
+        with open(demoddata.payload_frame.path) as frame_file:
+            # we get data frames in hex but kaitai requires binary
+            hexdata = frame_file.read()
+            bindata = binascii.unhexlify(hexdata)
+
+        try:
+            frame = decoder_class.from_bytes(bindata)
+            json_obj = create_point(
+                decoder.get_fields(frame), satellite, tlmdecoder, demoddata, decoders_version
+            )
+
+            # if we are set to use InfluxDB, send the decoded data
+            # there, otherwise we store it in the local DB.
+            if settings.USE_INFLUX:
+                write_influx(json_obj)
+                DemodData.objects.filter(pk=demoddata.id).update(
+                    is_decoded=True, payload_decoded='influxdb'
+                )
+            else:
+                DemodData.objects.filter(pk=demoddata.id).update(
+                    is_decoded=True, payload_decoded=json_obj
+                )
+        except Exception:  # pylint: disable=W0703
+            DemodData.objects.filter(pk=demoddata.id).update(is_decoded=False, payload_decoded='')
+    except (IOError, binascii.Error) as error:
+        LOGGER.error(error, exc_info=True)
+
+
+def decode_data(norad_id, demoddata_id=None, redecode=False):
     """Decode data for a satellite, with an option to limit the scope.
 
-    :param norad: the NORAD ID of the satellite to decode data for
-    :param period: if period exists, only attempt to decode the last 4 hours,
-    otherwise attempt to decode everything
-    :param handle: if handle exists, we are trying to decode the currently
-    saved frame that got in
+    :param norad_id: the NORAD ID of the satellite to decode data for
+    :param demoddata_id: if demoddata_id exists, try to decode this demoddata object
+    :param redecode: redecode demoddata or only recent
     """
-    sat = Satellite.objects.get(norad_cat_id=norad)
-    if not sat.telemetry_decoder_count:
-        return
-
-    if period:
-        time_period = datetime.utcnow() - timedelta(hours=4)
-        time_period = make_aware(time_period)
-        data = DemodData.objects.filter(satellite__norad_cat_id=norad, timestamp__gte=time_period)
-    else:
-        if handle:
-            data = DemodData.objects.filter(id=handle)
+    satellite = Satellite.objects.prefetch_related('telemetries').get(norad_cat_id=norad_id)
+    telemetry_decoders = satellite.telemetries.all()
+    if telemetry_decoders:
+        data = DemodData.objects.select_for_update()
+        if redecode:
+            if demoddata_id:
+                data = data.filter(pk=demoddata_id)
+            else:
+                data = data.filter(satellite__norad_cat_id=norad_id)
         else:
-            data = DemodData.objects.filter(satellite=sat)
-    telemetry_decoders = Telemetry.objects.filter(satellite=sat)
+            if demoddata_id:
+                data = data.filter(pk=demoddata_id, is_decoded=False)
+            else:
+                time_period = make_aware(datetime.utcnow() - timedelta(hours=48))
+                data = data.filter(
+                    satellite__norad_cat_id=norad_id, timestamp__gte=time_period, is_decoded=False
+                )
 
-    # iterate over DemodData objects
-    for obj in data:
-        # iterate over Telemetry decoders
-        for tlmdecoder in telemetry_decoders:
-            try:
-                decoder_class = getattr(decoder, tlmdecoder.decoder.capitalize())
-            except AttributeError:
-                continue
-            try:
-                with open(obj.payload_frame.path) as frame_file:
-                    # we get data frames in hex but kaitai requires binary
-                    hexdata = frame_file.read()
-                    bindata = binascii.unhexlify(hexdata)
-
-                # if we are set to use InfluxDB, send the decoded data
-                # there, otherwise we store it in the local DB.
-                if settings.USE_INFLUX:
-                    try:
-                        frame = decoder_class.from_bytes(bindata)
-                        json_obj = create_point(
-                            decoder.get_fields(frame), sat, tlmdecoder, obj, decoders_version
-                        )
-                        write_influx(json_obj)
-                        obj.payload_decoded = 'influxdb'
-                        obj.is_decoded = True
-                        obj.save()
-                        break
-                    except Exception:  # pylint: disable=W0703
-                        obj.is_decoded = False
-                        obj.save()
-                        continue
-                else:  # store in the local db instead of influx
-                    try:
-                        frame = decoder_class.from_bytes(bindata)
-                    except Exception:  # pylint: disable=W0703
-                        obj.payload_decoded = ''
-                        obj.is_decoded = False
-                        obj.save()
-                        continue
-                    else:
-                        json_obj = create_point(
-                            decoder.get_fields(frame), sat, tlmdecoder, obj, decoders_version
-                        )
-                        obj.payload_decoded = json_obj
-                        obj.is_decoded = True
-                        obj.save()
-                        break
-            except (IOError, binascii.Error) as error:
-                LOGGER.error(error, exc_info=True)
-                continue
+        with transaction.atomic():
+            # iterate over DemodData objects
+            for obj in data:
+                # iterate over Telemetry decoders
+                for tlmdecoder in telemetry_decoders:
+                    decode_demoddata(demoddata=obj, satellite=satellite, tlmdecoder=tlmdecoder)
 
 
 # Caches stats about satellites and data
